/*
 * Copyright 2016 - 2022 Gary Talent (gary@drinkingtea.net). All rights reserved.
 */

#pragma once

#include <ox/std/def.hpp>
#include <ox/std/error.hpp>
#include <ox/std/string.hpp>

#include <ox/claw/write.hpp>

#include "gfx.hpp"

namespace nostalgia::core {

class Wrap {
	public:
		virtual ~Wrap() = default;
};

template<typename T>
class WrapTemplate: public Wrap {
	private:
		T m_obj;

	public:
		constexpr WrapTemplate() = default;

		[[nodiscard]]
		constexpr auto obj() noexcept {
			return &m_obj;
		}

};

using WrapPtr = ox::UniquePtr<Wrap>;

template<typename T>
constexpr auto wrapCast(const auto &ptr) noexcept {
	return static_cast<WrapTemplate<T>*>(ptr.get())->obj();
}

struct BaseConverter {
	virtual ~BaseConverter() noexcept = default;

	[[nodiscard]]
	virtual const char *srcTypeName() noexcept = 0;

	[[nodiscard]]
	virtual int srcTypeVersion() noexcept = 0;

	[[nodiscard]]
	virtual bool srcMatches(const char *srcTypeName, int srcTypeVersion) const noexcept = 0;

	[[nodiscard]]
	virtual bool dstMatches(const char *dstTypeName, int dstTypeVersion) const noexcept = 0;

	virtual ox::Result<WrapPtr> convertPtrToPtr(const WrapPtr &src) noexcept = 0;

	virtual ox::Result<WrapPtr> convertBuffToPtr(const ox::Buffer &srcBuff) noexcept = 0;

	[[nodiscard]]
	inline bool matches(const char *srcTypeName, int srcTypeVersion,
	                    const char *dstTypeName, int dstTypeVersion) const noexcept {
		return srcMatches(srcTypeName, srcTypeVersion)
		    && dstMatches(dstTypeName, dstTypeVersion);
	}

};

template<typename SrcType, typename DstType>
struct Converter: public BaseConverter{

	virtual ox::Error convert(SrcType*, DstType*) noexcept = 0;

	[[nodiscard]]
	const char *srcTypeName() noexcept final {
		return SrcType::TypeName;
	}

	[[nodiscard]]
	int srcTypeVersion() noexcept final {
		return SrcType::TypeVersion;
	}

	[[nodiscard]]
	bool srcMatches(const char *srcTypeName, int srcTypeVersion) const noexcept final {
		return ox_strcmp(srcTypeName, SrcType::TypeName) == 0
		    && srcTypeVersion == SrcType::TypeVersion;
	}

	[[nodiscard]]
	bool dstMatches(const char *dstTypeName, int dstTypeVersion) const noexcept final {
		return ox_strcmp(dstTypeName, DstType::TypeName) == 0
		    && dstTypeVersion == DstType::TypeVersion;
	}

	ox::Result<WrapPtr> convertPtrToPtr(const WrapPtr &src) noexcept final {
		auto dst = ox::make_unique<WrapTemplate<DstType>>();
		oxReturnError(convert(wrapCast<SrcType>(src), wrapCast<DstType>(dst)));
		return ox::Result<WrapPtr>(std::move(dst));
	}

	ox::Result<WrapPtr> convertBuffToPtr(const ox::Buffer &srcBuff) noexcept final {
		oxRequireM(src, ox::readClaw<SrcType>(srcBuff));
		auto dst = ox::make_unique<WrapTemplate<DstType>>();
		oxReturnError(convert(&src, wrapCast<DstType>(dst)));
		return ox::Result<WrapPtr>(std::move(dst));
	}

};

ox::Result<WrapPtr> convert(const ox::Buffer &srcBuffer, const char *dstTypeName, int dstTypeVersion) noexcept;

template<typename DstType>
ox::Result<DstType> convert(const ox::Buffer &srcBuffer) noexcept {
	oxRequire(out, convert(srcBuffer, DstType::TypeName, DstType::TypeVersion));
	return wrapCast<DstType>(out);
}

template<typename T>
ox::Error convert(const ox::Buffer &buff, T *outObj) noexcept {
	oxRequire(outPtr, convert(buff, T::TypeName, T::TypeVersion));
	*outObj = std::move(*wrapCast<T>(outPtr));
	return OxError(0);
}

template<typename DstType>
ox::Result<ox::Buffer> convertBuffToBuff(const ox::Buffer &srcBuffer, ox::ClawFormat fmt) noexcept {
	oxRequire(out, convert(srcBuffer, DstType::TypeName, DstType::TypeVersion));
	return ox::writeClaw<DstType>(wrapCast<DstType>(out), fmt);
}

}
